import argparse
import numpy as np
import filecmp
from shutil import copyfile
from functools import partial
from itertools import tee, count

from .utils import tsv_reader, write_matrix_and_index_file, determine_index_file_name


def create_index_map(index_file):
    index_map = dict()
    with open(index_file) as fp:
        for line in tsv_reader(fp):
            if line[0].startswith("#"):
                continue
            index_map[line[1]] = int(line[0])
    return index_map


def pick_subset_by_row_index(X, index_pairs, n):
    _, m = X.shape
    A = np.zeros((n, m), dtype=X.dtype)
    for i, j in index_pairs:
        A[i] = X[j]
    return A


def unzip_pair_iter(it):
    it1, it2 = tee(it)
    first_it = (a for a, b in it1)
    second_it = (b for a, b in it2)
    return (first_it, second_it)


def get_id_index_pair_iters_from_ids(index_map, ids):

    pairing = lambda index_map, row_id: (row_id, index_map[row_id])
    
    pair_iter = map(partial(pairing, index_map), ids)
    ids_iter, index_iter = unzip_pair_iter(pair_iter)
    new_index_id_pairs = zip(count(), ids_iter)
    new_old_index_pairs = zip(count(), index_iter)
    return (new_index_id_pairs, new_old_index_pairs)


def pick_subset(matrix_file, index_file, output_file, ids):
    """ Given a matrix and a list of row identifiers, create a 
        new matrix which is a row-wise subset of the given matrix.

        :param matrix_file: File path to the input matrix
        :type matrix_file: str
        :param index_file: File path to the index file which coresponds to the input matrix
        :type index_file: str
        :param output_file: File path to the output matrix
        :type output_file: str
        :param ids: List of row identifiers for the input matrix
        :type ids: List[str]
        :returns:  None
    """
    matrix = np.load(matrix_file)
    index_map = create_index_map(index_file)
    new_index_id_pairs, new_old_index_pairs = get_id_index_pair_iters_from_ids(
        index_map, ids
    )
    sub_matrix = pick_subset_by_row_index(matrix, new_old_index_pairs, len(ids))
    write_matrix_and_index_file(output_file, sub_matrix, new_index_id_pairs)


def create_ids_map(index_file):
    ids_map = dict()
    with open(index_file) as fp:
        for line in tsv_reader(fp):
            if line[0].startswith("#"):
                continue
            ids_map[int(line[0])] = line[1]
    return ids_map


def row_cov_pair(A, i):
    return (i, np.sum(A[i]))


def row_coverage_pairs(A):
    n, _ = A.shape
    return map(partial(row_cov_pair, A), range(n))


first = lambda pair: pair[0]
second = lambda pair: pair[1]


def pair_inside_limits(lower, upper, pair):
    return lower < second(pair) < upper


def get_id_index_pair_iters_from_indexs(ids_map, indexs):

    pairing = lambda ids_map, index: (ids_map[index], index)

    pair_iter = map(partial(pairing, ids_map), indexs)
    ids_iter, index_iter = unzip_pair_iter(pair_iter)
    new_index_id_pairs = zip(count(), ids_iter)
    new_old_index_pairs = zip(count(), index_iter)
    return (new_index_id_pairs, new_old_index_pairs)


def cut_tails(limit_func, matrix, index_file, output_file, cut):
    A = np.load(matrix)
    row_cov_pairs = sorted(row_coverage_pairs(A), key=second)
    lower, upper = limit_func(row_cov_pairs, cut)
    subset_indexs = list(map(
        first, filter(partial(pair_inside_limits, lower, upper), row_cov_pairs)
    ))
    ids_map = create_ids_map(index_file)
    new_index_id_pairs, new_old_index_pairs = get_id_index_pair_iters_from_indexs(
        ids_map, subset_indexs
    )
    sub_matrix = pick_subset_by_row_index(
        A, new_old_index_pairs, len(subset_indexs)
    )
    write_matrix_and_index_file(output_file, sub_matrix, new_index_id_pairs)


def find_limits_double_cut(row_cov_pairs, cut):
    cut_index = int(len(row_cov_pairs) * cut)
    lower = second(row_cov_pairs[cut_index])
    upper = second(row_cov_pairs[-cut_index])
    return (lower, upper)


def find_limits_left_cut(row_cov_pairs, cut):
    cut_index = int(len(row_cov_pairs) * cut)
    lower = second(row_cov_pairs[cut_index])
    upper = float("inf")
    return (lower, upper)


def find_limits_right_cut(row_cov_pairs, cut):
    cut_index = int(len(row_cov_pairs) * cut)
    lower = -1
    upper = second(row_cov_pairs[-cut_index])
    return (lower, upper)


def cut_tails_both(matrix, index_file, output_file, cut=0.05):
    """ Removes the left and right tail from the coverage distribution of the
        matrix. The cuts based on a fraction given as cut.
        If cut = 0.05 the top 5% and the bottom 5% will be removed.

        :param matrix_file: File path to the input matrix
        :type matrix_file: str
        :param index_file: File path to the index file which coresponds to the input matrix
        :type index_file: str
        :param output_file: File path to the output matrix
        :type output_file: str
        :param cut: Fraction to be removed
        :type cut: 0 < int < 1
        :returns:  None
    """
    cut_tails(find_limits_double_cut, matrix, index_file, output_file, cut)


def cut_tails_left(matrix, index_file, output_file, cut=0.05):
    """ Removes the left tail from the coverage distribution of the matrix. 
        The cut is based on a fraction given as cut.
        If cut = 0.05 the bottom 5% will be removed.

        :param matrix_file: File path to the input matrix
        :type matrix_file: str
        :param index_file: File path to the index file which coresponds to the input matrix
        :type index_file: str
        :param output_file: File path to the output matrix
        :type output_file: str
        :param cut: Fraction to be removed
        :type cut: 0 < int < 1
        :returns:  None
    """
    cut_tails(find_limits_left_cut, matrix, index_file, output_file, cut)


def cut_tails_right(matrix, index_file, output_file, cut=0.05):
    """ Removes the right tail from the coverage distribution of the matrix. 
        The cut is based on a fraction given as cut.
        If cut = 0.05 the top 5% will be removed.

        :param matrix_file: File path to the input matrix
        :type matrix_file: str
        :param index_file: File path to the index file which coresponds to the input matrix
        :type index_file: str
        :param output_file: File path to the output matrix
        :type output_file: str
        :param cut: Fraction to be removed
        :type cut: 0 < int < 1
        :returns:  None
    """
    cut_tails(find_limits_right_cut, matrix, index_file, output_file, cut)


def first_true(pred_func, it):
    for elm in it:
        if pred_func(elm):
            return elm


def assert_index_files_are_identical(index_files):
    not_none = lambda x: x != None
    ground_truth = first_true(not_none, index_files)
    if ground_truth == None:
        print("Warning: No index files found")
        return None
    filtered_index_files = list(filter(not_none, index_files))
    if len(index_files) != len(filtered_index_files):
        print("Warning: Index files missing")
    assert all(
        map(partial(filecmp.cmp, ground_truth, shallow=False), filtered_index_files)
    ), "Index files doesn't match"
    return ground_truth


def sum_matrices(matrices, start, end, uint32):
    summary_matrix = np.load(matrices[0])
    if end == None:
        _, end = summary_matrix.shape
    summary_matrix = summary_matrix[:, start:end]
    if uint32 == True:
        dtype = np.uint32
        summary_matrix = summary_matrix.astype(dtype)
    else:
        dtype = summary_matrix.dtype
    if len(matrices) > 1:
        for input_file in matrices[1:]:
            A = np.load(input_file).astype(dtype)[:, start:end]
            summary_matrix += A
    return summary_matrix


def collapse(matrix_index_pairs, output_file, start=0, end=None, uint32=False):
    """ This function will given a list matrix-index pairs, load in the matrices one by
        one and collapse them upon each other value by value.
        The final matrix will be stored in the output file as a .npy file, with a 
        corresponding index file.

        Use start and end if you want to only collapse and output a certain region.

        :param matrix_index_pairs: List of pairs (file path to matrix, file path index file)
        :type matrices: List((str,str))
        :param output_file: File path to the output file
        :type output_file: str
        :param start: start index for collapsing
        :type start: int >= 0
        :param end: end index for collapsing, if None then last index is end.
        :type end: int >= 0
        :param uint32: If False numpy.dtype will be preserved, if True numpy.dtype will be changed to uint32
        :type uint32: boolean
        :returns:  None
    """
    matrix_files, index_files = list(zip(*matrix_index_pairs))
    index_file = assert_index_files_are_identical(index_files)
    summary_matrix = sum_matrices(matrix_files, start, end, uint32)
    copyfile(index_file, determine_index_file_name(output_file))
    np.save(output_file, summary_matrix)



import unittest
import numpy as np
import tempfile
import filecmp
from os.path import join, dirname

import transcov.manipulations as mut


class Test_pick_subset_by_row_index(unittest.TestCase):
    def setUp(self):
        self.X = np.arange(12, dtype=np.uint16).reshape(4, 3)

    def test_pick_two_rows(self):
        index_pairs = ((0, 1), (1, 2))
        n = len(index_pairs)
        res = mut.pick_subset_by_row_index(self.X, index_pairs, n)
        exp_res = np.array([[3, 4, 5], [6, 7, 8]], dtype=np.uint16)
        self.assertTrue((res == exp_res).all())


class Test_pick_subset_integration(unittest.TestCase):
    def setUp(self):
        # create input matrix
        self.input_matrix = tempfile.NamedTemporaryFile()
        A = np.arange(12, dtype=np.uint16).reshape(4, 3)
        np.save(self.input_matrix.name, A)
        # create input index
        self.input_index = tempfile.NamedTemporaryFile()
        with open(self.input_index.name, "w") as fp:
            fp.write("#index\tid\n0\ta\n1\tb\n2\tc\n3\td")
        # instantiate output file
        self.output_matrix = tempfile.NamedTemporaryFile()

    def tearDown(self):
        self.input_matrix.close()
        self.input_index.close()
        self.output_matrix.close()

    def test_pick_subset(self):
        ids = ["b", "c"]
        mut.pick_subset(
            self.input_matrix.name + ".npy",
            self.input_index.name,
            self.output_matrix.name,
            ids,
        )
        exp_res = np.array([[3, 4, 5], [6, 7, 8]], dtype=np.uint16)
        res = np.load(self.output_matrix.name + ".npy")
        self.assertTrue((res == exp_res).all())

class Test_collapse_integration(unittest.TestCase):
    def setUp(self):
        self.output_matrix = tempfile.NamedTemporaryFile()

    def tearDown(self):
        self.output_matrix.close()

    def test_coverage_collapse(self):
        matrix = join(dirname(__file__), "data/coverage_matrix.npy")
        index = join(dirname(__file__), "data/coverage_matrix.index")
        matrix_index_pairs = [(matrix, index), (matrix, index)]
        mut.collapse(matrix_index_pairs, self.output_matrix.name)
        A = np.load(matrix)
        B = np.load(self.output_matrix.name + ".npy")
        self.assertTrue(((A+A)==B).all())

class Test_cut_tails_both_integration(unittest.TestCase):
    def setUp(self):
        self.output_matrix = tempfile.NamedTemporaryFile()

    def tearDown(self):
        self.output_matrix.close()

    def test_cut_tails_both(self):
        matrix = join(dirname(__file__), "data/coverage_matrix.npy")
        index = join(dirname(__file__), "data/coverage_matrix.index")
        mut.cut_tails_both(matrix, index, self.output_matrix.name)
        An, Am = np.load(matrix).shape
        Bn, Bm = np.load(self.output_matrix.name + ".npy").shape
        self.assertTrue(An>Bn)
        self.assertTrue(Am==Bm)

if __name__ == "__main__":
    unittest.main()



from .generator import (
    generate_coverage_matrix,
    generate_read_ends_matrix,
    generate_length_matrix,
)
from .preprocessor import preprocess
from .manipulations import collapse, pick_subset, cut_tails_left, cut_tails_right, cut_tails_both

__version__ = "1.0.6"

__all__ = (
    "generate_coverage_matrix",
    "generate_read_ends_matrix",
    "generate_length_matrix",
    "preprocess",
    "collapse",
    "pick_subset",
    "cut_tails_both",
    "cut_tails_right",
    "cut_tails_both",
)



import unittest
import numpy as np
import tempfile
import filecmp
from os.path import join, dirname

from transcov import generator


class Test_calc_rel_start_and_end_Function(unittest.TestCase):
    def setUp(self):
        self.tss = 1000

    def test_total_lower_pos_strand(self):
        start = 550
        end = 770
        strand = "+"
        res_start, res_end = generator.calc_rel_start_and_end(
            start, end, strand, self.tss
        )
        self.assertEqual(res_start, -450)
        self.assertEqual(res_end, -230)

    def test_total_upper_pos_strand(self):
        start = 1300
        end = 1450
        strand = "+"
        res_start, res_end = generator.calc_rel_start_and_end(
            start, end, strand, self.tss
        )
        self.assertEqual(res_start, 300)
        self.assertEqual(res_end, 450)

    def test_total_upper_neg_strand(self):
        start = 1300
        end = 1450
        strand = "-"
        res_start, res_end = generator.calc_rel_start_and_end(
            start, end, strand, self.tss
        )
        self.assertEqual(res_start, -450)
        self.assertEqual(res_end, -300)

    def test_overlap_neg_strand(self):
        start = 600
        end = 1450
        strand = "-"
        res_start, res_end = generator.calc_rel_start_and_end(
            start, end, strand, self.tss
        )
        self.assertEqual(res_start, -450)
        self.assertEqual(res_end, 400)


class Test_add_read_ends_Function(unittest.TestCase):
    def setUp(self):
        self.A = np.zeros((3, 101), dtype=np.uint16)
        self.k = 50

    def test_add_center_fragment(self):
        start = -30
        end = 30
        i = 1
        generator.add_read_ends(self.A, start, end, i, self.k)
        res = np.zeros(101, dtype=np.uint16)
        res[start + self.k] = 1
        res[end + self.k] = 1
        self.assertTrue((self.A[i] == res).all())

    def test_left_out_of_bounds_fragment(self):
        start = -300
        end = 30
        i = 0
        generator.add_read_ends(self.A, start, end, i, self.k)
        res = np.zeros(101, dtype=np.uint16)
        res[end + self.k] = 1
        self.assertTrue((self.A[i] == res).all())

    def test_addtion_side_effect_fragment(self):
        start = -300
        end = 30
        i = 1
        generator.add_read_ends(self.A, start, end, i, self.k)
        generator.add_read_ends(self.A, start, end, i, self.k)
        res = np.zeros(101, dtype=np.uint16)
        res[end + self.k] = 2
        self.assertTrue((self.A[i] == res).all())

    def test_edges_fragment(self):
        start = -50
        end = 50
        i = 2
        generator.add_read_ends(self.A, start, end, i, self.k)
        res = np.zeros(101, dtype=np.uint16)
        res[start + self.k] = 1
        res[end + self.k] = 1
        self.assertTrue((self.A[i] == res).all())


class Test_add_fragment_Function(unittest.TestCase):
    def setUp(self):
        self.A = np.zeros((3, 101), dtype=np.uint16)
        self.k = 50

    def test_add_center_fragment(self):
        start = -30
        end = 30
        i = 1
        generator.add_fragment(self.A, start, end, i, self.k)
        res = np.zeros(101, dtype=np.uint16)
        res[start + self.k : end + self.k] = 1
        self.assertTrue((self.A[i] == res).all())

    def test_left_out_of_bounds_fragment(self):
        start = -300
        end = 30
        i = 0
        generator.add_fragment(self.A, start, end, i, self.k)
        res = np.zeros(101, dtype=np.uint16)
        res[start + self.k : end + self.k] = 1
        self.assertTrue((self.A[i] == res).all())

    def test_addtion_side_effect_fragment(self):
        start = -300
        end = 30
        i = 1
        generator.add_fragment(self.A, start, end, i, self.k)
        generator.add_fragment(self.A, start, end, i, self.k)
        res = np.zeros(101, dtype=np.uint16)
        res[start + self.k : end + self.k] = 2
        self.assertTrue((self.A[i] == res).all())

    def test_edges_fragment(self):
        start = -50
        end = 51
        i = 2
        generator.add_fragment(self.A, start, end, i, self.k)
        res = np.ones(101, dtype=np.uint16)
        self.assertTrue((self.A[i] == res).all())


class Test_module_with_file_io(unittest.TestCase):
    def setUp(self):
        self.output_matrix = tempfile.NamedTemporaryFile()

    def tearDown(self):
        self.output_matrix.close()

    def test_read_ends_matrix_generation(self):
        bam_file = join(dirname(__file__), "data/test.bam")
        bed_file = join(dirname(__file__), "data/gencode.v19.annotation.tss.bed")
        res_matrix = join(dirname(__file__), "data/read_ends_matrix.npy")
        generator.generate_read_ends_matrix(bam_file, bed_file, self.output_matrix.name)
        res = filecmp.cmp(res_matrix, self.output_matrix.name + ".npy", shallow=False)
        self.assertTrue(res)

    def test_coverage_matrix_generation(self):
        bam_file = join(dirname(__file__), "data/test.bam")
        bed_file = join(dirname(__file__), "data/gencode.v19.annotation.tss.bed")
        res_matrix = join(dirname(__file__), "data/coverage_matrix.npy")
        generator.generate_coverage_matrix(bam_file, bed_file, self.output_matrix.name)
        res = filecmp.cmp(res_matrix, self.output_matrix.name + ".npy", shallow=False)
        self.assertTrue(res)

    def test_length_matrix_generation(self):
        bam_file = join(dirname(__file__), "data/test.bam")
        bed_file = join(dirname(__file__), "data/gencode.v19.annotation.tss.bed")
        res_matrix = join(dirname(__file__), "data/length_matrix.npy")
        generator.generate_length_matrix(bam_file, bed_file, self.output_matrix.name)
        res = filecmp.cmp(res_matrix, self.output_matrix.name + ".npy", shallow=False)
        self.assertTrue(res)


if __name__ == "__main__":
    unittest.main()



import click
from os.path import isfile

from .generator import (
    generate_coverage_matrix,
    generate_read_ends_matrix,
    generate_length_matrix,
)
from . import preprocessor
from . import manipulations
from .utils import tsv_reader, determine_index_file_name


@click.group()
def cli():
    # Add debug info like:
    # click.echo('No matrices was given to collapse')
    # click.echo('Only one matrix was given to collapse')
    pass


@cli.command()
@click.argument("annotation_file")
@click.option("-k", "--region-size", default=10000, type=click.IntRange(min=0))
@click.option("--bed-file", default="transcription_start_sites.bed")
@click.option("--tss-file", default="transcription_start_sites.tsv")
def preprocess(annotation_file, region_size, bed_file, tss_file):
    preprocessor.preprocess(annotation_file, region_size, bed_file, tss_file)


@cli.command()
@click.argument("bam_file")
@click.argument("bed_file")
@click.option("-o", "--output-file", default="coverage_matrix.npy")
def generate_coverage(bam_file, bed_file, output_file):
    generate_coverage_matrix(bam_file, bed_file, output_file)


@cli.command()
@click.argument("bam_file")
@click.argument("bed_file")
@click.option("-o", "--output-file", default="read_ends_matrix.npy")
def generate_read_ends(bam_file, bed_file, output_file):
    generate_read_ends_matrix(bam_file, bed_file, output_file)


@cli.command()
@click.argument("bam_file")
@click.argument("bed_file")
@click.option("-o", "--output-file", default="length_matrix.npy")
@click.option("-m", "--max-length", default=500, type=click.IntRange(min=1))
def generate_length(bam_file, bed_file, output_file, max_length):
    generate_length_matrix(bam_file, bed_file, output_file, max_length)


@cli.command()
@click.argument("matrices", nargs=-1)
@click.option("-o", "--output-file", default="collapsed_matrix.npy")
@click.option("-s", "--start", default=0, type=click.IntRange(min=0))
@click.option("-e", "--end", default=None, type=int)
@click.option("--uint32", is_flag=True)
def collapse(matrices, output_file, start, end, uint32):
    if len(matrices) > 0:

        def isfile_or_none(file_path):
            if isfile(file_path):
                return file_path
            else:
                return None

        indexes = map(isfile_or_none, map(determine_index_file_name, matrices))
        print(indexes)
        manipulations.collapse(zip(matrices, indexes), output_file, start, end, uint32)


@cli.command()
@click.argument("input_matrix")
@click.argument("index_file")
@click.argument("ids_file")
@click.option("-o", "--output-file", default="subset_matrix.npy")
def pick_subset(input_matrix, index_file, ids_file, output_file):
    ids = list()
    with open(ids_file) as fp:
        for line in tsv_reader(fp):
            if line[0].startswith("#"):
                continue
            ids.append(line[0])
    manipulations.pick_subset(input_matrix, index_file, output_file, ids)


@cli.command()
@click.argument("input_matrix")
@click.argument("index_file")
@click.option("-o", "--output-file", default="tail_cutted_matrix.npy")
@click.option("-c", "--cut", default=0.05, type=click.FloatRange(min=0,max=1))
@click.option("-m", "--mode", default="both", type=click.Choice(['both','left','right'], case_sensitive=False))
def cut_tails(input_matrix, index_file, output_file, cut, mode):
    tail_funcs = {'both': manipulations.cut_tails_both,
                  'left': manipulations.cut_tails_left,
                  'right': manipulations.cut_tails_right,
                  }
    tail_funcs[mode](input_matrix, index_file, output_file, cut)



import attr
import csv


@attr.s
class TranscriptionStartSite:
    tss_id = attr.ib()
    chrom = attr.ib()
    tss = attr.ib()
    strand = attr.ib()
    tx_ids = attr.ib()
    gene_id = attr.ib()
    ccds_id = attr.ib()
    tx_types = attr.ib()

    def __str__(self):
        return "\t".join(
            (
                self.tss_id,
                self.chrom,
                str(self.tss),  # is integer
                self.strand,
                " ".join(self.tx_ids),
                self.gene_id,
                str(self.ccds_id),  # is string or None
                " ".join(self.tx_types),
            )
        )


def get_header():
    return "\t".join(
        ("tss_id", "chrom", "tss", "strand", "tx_ids", "gene_id", "ccds_id", "tx_types")
    )


def load_transcription_start_sites(input_file):
    TSSs = list()
    with open(input_file) as fp:
        tsv_reader = csv.reader(fp, delimiter="\t")
        header = next(tsv_reader)
        for line in tsv_reader:
            tss = TranscriptionStartSite(
                line[0],
                line[1],
                int(line[2]),
                line[3],
                line[4].split(),
                line[5],
                line[6],
                line[7].split(),
            )
            TSSs.append(tss)
    return TSSs



import csv
import numpy as np


def tsv_writer(fp):
    """ Given a file obj return csv.writer with delimiter '\t' """
    return csv.writer(fp, delimiter="\t")


def tsv_reader(fp):
    """ Given a file obj return csv.reader with delimiter '\t' """
    return csv.reader(fp, delimiter="\t")


def determine_index_file_name(output_file):
    if output_file.split(".")[-1] == "npy":
        return output_file.replace(".npy", ".index")
    else:
        return output_file + ".index"


def write_index_file(output_file, index_lst):
    with open(determine_index_file_name(output_file), "w") as fp:
        writer = tsv_writer(fp)
        writer.writerow(("#index", "id"))
        for row_info in index_lst:
            writer.writerow(row_info)


def write_matrix_and_index_file(output_file, matrix, index_lst):
    write_index_file(output_file, index_lst)
    np.save(output_file, matrix)



import numpy as np
import csv
import attr

from .bam import BAM
from .utils import write_matrix_and_index_file


@attr.s
class BED:
    chrom = attr.ib()
    start = attr.ib()
    end = attr.ib()
    tss_id = attr.ib()
    score = attr.ib()
    strand = attr.ib()


def load_bed_file(file_path):
    bed_list = list()
    with open(file_path) as fp:
        reader = csv.reader(fp, delimiter="\t")
        for line in reader:
            if line[0].startswith("#"):
                continue
            bed_list.append(
                BED(line[0], int(line[1]), int(line[2]), line[3], int(line[4]), line[5])
            )
    return bed_list


def calc_rel_start_and_end(read_start, read_end, strand, tss):
    rel_start = read_start - tss
    rel_end = read_end - tss
    if strand == "+":
        return (rel_start, rel_end)
    elif strand == "-":
        return (-rel_end, -rel_start)


def add_read_ends(A, start, end, i, k):
    _, m = A.shape
    a = start + k
    b = end + k
    if a >= 0 and a < m:
        A[i, a] += 1
    if b >= 0 and b < m:
        A[i, b] += 1


def add_fragment(A, start, end, i, k):
    _, m = A.shape
    a = min((start, end)) + k
    b = max((start, end)) + k
    if a < 0:
        a = 0
    if b > m:
        b = m
    v = np.zeros(m, dtype=A.dtype)
    v[a:b] = 1
    A[i] += v


def generate_length_matrix(bam_file, bed_file, output_file, max_length=500):
    """ Creates a matrix where each row represents a region from the bed file
        and the columns are read lengths from 0 to max_length.
        The size of the matrix is (n x max_length) where n is the number of regions
        in the bed file.
        Data is read length counts, so that a_ij is the number of reads in region i, 
        with length j.

        :param bam_file: File path to the bam sample file
        :type bam_file: str
        :param bed_file: File path to the bed file, which can be compiled by the preprocessing function
        :type bed_file: str
        :param output_file: File path to the output file
        :type output_file: str
        :param max_length: Maximum read length to be counted
        :type max_length: int > 0
        :returns:  None
    """
    bed_list = load_bed_file(bed_file)
    matrix = np.zeros((len(bed_list), max_length), dtype=np.uint16)
    bam = BAM(bam_file)
    index_lst = list()
    for i, region in enumerate(bed_list):
        for reading in bam.pair_generator(region.chrom, region.start, region.end):
            start = int(reading[1])
            end = int(reading[2])
            length = abs(end - start)
            if length < max_length:
                matrix[i, length] += 1
        index_lst.append((i, region.tss_id))
    write_matrix_and_index_file(output_file, matrix, index_lst)


def generate_read_ends_matrix(bam_file, bed_file, output_file):
    """ Creates a matrix where each row represents a region with a TSS in the
        center of the region and the columns represents base pair positions.
        The size of the matrix is (n x m) where n is the number of regions/TSS
        and m is the size og the region.
        Data is read-end counts, so that a_ij is the number of read-ends in region
        i at position j from region start.

        :param bam_file: File path to the bam sample file
        :type bam_file: str
        :param bed_file: File path to the bed file, which can be compiled by the preprocessing function
        :type bed_file: str
        :param output_file: File path to the output file
        :type output_file: str
        :returns:  None
    """
    bed_list = load_bed_file(bed_file)
    region_size = bed_list[0].end - bed_list[0].start
    coverage_matrix = np.zeros((len(bed_list), region_size), dtype=np.uint16)
    bam = BAM(bam_file)
    index_lst = list()
    for i, region in enumerate(bed_list):
        tss = int(region.tss_id.split("_")[-1])
        for reading in bam.pair_generator(region.chrom, region.start, region.end):
            read_start = int(reading[1])
            read_end = int(reading[2])
            rel_start, rel_end = calc_rel_start_and_end(
                read_start, read_end, region.strand, tss
            )
            k = tss - region.start
            add_read_ends(coverage_matrix, rel_start, rel_end, i, k)
        index_lst.append((i, region.tss_id))
    write_matrix_and_index_file(output_file, coverage_matrix, index_lst)


def generate_coverage_matrix(bam_file, bed_file, output_file):
    """ Creates a matrix where each row represents a region with a TSS in the
        center of the region and the columns represents base pair positions.
        The size of the matrix is (n x m) where n is the number of regions/TSS
        and m is the size og the region.
        Data is read-depth/coverage, so that a_ij is the number of overlapping 
        reads in region i at position j from region start.

        :param bam_file: File path to the bam sample file
        :type bam_file: str
        :param bed_file: File path to the bed file, which can be compiled by the preprocessing function
        :type bed_file: str
        :param output_file: File path to the output file
        :type output_file: str
        :returns:  None
    """
    bed_list = load_bed_file(bed_file)
    region_size = bed_list[0].end - bed_list[0].start
    coverage_matrix = np.zeros((len(bed_list), region_size), dtype=np.uint16)
    bam = BAM(bam_file)
    index_lst = list()
    for i, region in enumerate(bed_list):
        tss = int(region.tss_id.split("_")[-1])
        for reading in bam.pair_generator(region.chrom, region.start, region.end):
            read_start = int(reading[1])
            read_end = int(reading[2])
            rel_start, rel_end = calc_rel_start_and_end(
                read_start, read_end, region.strand, tss
            )
            k = tss - region.start
            add_fragment(coverage_matrix, rel_start, rel_end, i, k)
        index_lst.append((i, region.tss_id))
    write_matrix_and_index_file(output_file, coverage_matrix, index_lst)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="transcov",
    version="1.0.6",
    author="Per Høgfeldt",
    description="A software for mapping coverage around transcription start sites",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hogfeldt/transcov",
    packages=find_packages("src"),
    package_dir={"": "src"},
    test_suite="test",
    install_requires=[
        "pysam",  # 0.15.4
        "numpy",  # 1.18.1
        "attrs",  # 19.3.0
        "click>=6.0",
    ],
    entry_points={"console_scripts": ["transcov = transcov.cli:cli"],},
    include_package_data=True,
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)



import argparse
import attr
import re
import csv

from .transcript_annotation import (
    pull_tx_id,
    pull_ensemble_gene_id,
    pull_ccds_id,
    pull_tx_type,
)
from .tss import TranscriptionStartSite, get_header


@attr.s
class Tx_annotation:
    chrom = attr.ib()
    start = attr.ib()
    end = attr.ib()
    strand = attr.ib()
    additionals = attr.ib()


def is_chrome_autosome(chrom):
    return re.match(r"chr\d", chrom) != None


def get_transcript_annotations(file_path):
    with open(file_path, "r") as file_obj:
        for line in file_obj:
            if line[0] == "#":  # line is comment
                continue
            anno = line.replace("\n", "").split()
            if anno[2] != "transcript":  # line is not a transcript
                continue
            if is_chrome_autosome(anno[0]):
                anno_obj = Tx_annotation(
                    anno[0], int(anno[3]), int(anno[4]), anno[6], anno[8]
                )
                yield anno_obj


def determine_TSS_and_format_data(tx_anno):
    if tx_anno.strand == "+":
        TSS = tx_anno.start
    elif tx_anno.strand == "-":
        TSS = tx_anno.end
    else:
        raise Exception("Strand annotation is neither '+' nor '-'")
    tss_id = "_".join((tx_anno.chrom, str(TSS)))
    return TranscriptionStartSite(
        tss_id,
        tx_anno.chrom,
        TSS,
        tx_anno.strand,
        [pull_tx_id(tx_anno)],
        pull_ensemble_gene_id(tx_anno),
        pull_ccds_id(tx_anno),
        [pull_tx_type(tx_anno)],
    )


def preprocess(input_file, region_size, bed_file, tss_file):
    """ This function will given a gencode annotation file, find all transcripts
        and determine the Transcription Start Site (TSS) for the transcript.
        Information about the TSS will be stored in the tss file with metadata
        and a bed file which can be used as input for the generator.

        :param input_file: File path to the gencode annotation file
        :type input_file: str
        :param region_size: Size of the region with the TSS in the center which
                            should be stored in the bed file
        :type region_size: int >= 0
        :param bed_file: File path to the bed file
        :type bed_file: str
        :param tss_file: File path to the tss file
        :type tss_file: str
        :returns:  None
    """
    tx_annotations = get_transcript_annotations(input_file)
    TSS_dict = dict()
    for tx_anno in tx_annotations:
        tss = determine_TSS_and_format_data(tx_anno)
        if tss.tss_id in TSS_dict:
            if tss.tx_types[0] not in TSS_dict[tss.tss_id].tx_types:
                TSS_dict[tss.tss_id].tx_types += tss.tx_types
            TSS_dict[tss.tss_id].tx_ids += tss.tx_ids
        else:
            TSS_dict[tss.tss_id] = tss
    with open(bed_file, "w") as fp:
        k = int(region_size / 2)
        writer = csv.writer(fp, delimiter="\t")
        writer.writerow(["#chrom", "start", "end", "name", "score", "strand"])
        for tss_id in TSS_dict.keys():
            tss = TSS_dict[tss_id]
            writer.writerow(
                [tss.chrom, tss.tss - k, tss.tss + k, tss_id, 0, tss.strand]
            )
    with open(tss_file, "w") as fp:
        fp.write("#%s\n" % get_header())
        for tss_id in TSS_dict.keys():
            fp.write("%s\n" % str(TSS_dict[tss_id]))
    return (bed_file, tss_file)
